import sys
import os
import json
import comtypes
from ctypes import cast, POINTER
from comtypes import CLSCTX_ALL
from pycaw.pycaw import AudioUtilities, IAudioEndpointVolume
import soundcard as sc
import soundfile as sf
import time
import librosa
import numpy as np
import tkinter as tk
from tkinter import messagebox
import threading
import pystray
from PIL import Image, ImageDraw

VERSION = 'V1.0'

# Valeurs par défaut
DEFAULT_LIMIT_DB = 80
DEFAULT_WAIT_TIME = 15  # Durée d'attente par défaut en secondes

LIMIT_DB = DEFAULT_LIMIT_DB
SAMPLE_RATE = 48000
RECORD_SEC = 3

running = False  # Variable globale pour contrôler l'état du programme
icon = None  # Variable pour l'icône de la systray

def resource_path(relative_path):
    """ Get absolute path to resource, works for dev and for PyInstaller """
    try:
        base_path = sys._MEIPASS
    except Exception:
        base_path = os.path.abspath(".")
    return os.path.join(base_path, relative_path)

def load_settings():
    """Charge les paramètres à partir d'un fichier JSON."""
    global LIMIT_DB, WAIT_TIME
    LIMIT_DB = DEFAULT_LIMIT_DB
    WAIT_TIME = DEFAULT_WAIT_TIME
    try:
        with open(resource_path(r"data\settings.json"), 'r') as f:
            settings = json.load(f)
            LIMIT_DB = settings.get('LIMIT_DB', DEFAULT_LIMIT_DB)
            WAIT_TIME = settings.get('WAIT_TIME', DEFAULT_WAIT_TIME)
    except FileNotFoundError:
        print("Aucun fichier de paramètres trouvé, utilisation des valeurs par défaut.")
    except json.JSONDecodeError:
        print("Erreur lors du chargement du fichier de paramètres.")

def save_settings():
    """Enregistre les paramètres dans un fichier JSON."""
    settings = {
        'LIMIT_DB': LIMIT_DB,
        'WAIT_TIME': WAIT_TIME
    }
    with open(resource_path(r"data\settings.json"), 'w') as f:
        json.dump(settings, f)

def record_file():
    print('Recording...')
    with sc.get_microphone(id=str(sc.default_speaker().name), include_loopback=True).recorder(samplerate=SAMPLE_RATE) as mic:
        data = mic.record(numframes=SAMPLE_RATE * RECORD_SEC)
        sf.write(file=resource_path(r'data\out.wav'), data=data[:, 0], samplerate=SAMPLE_RATE)
    print('Stop Recording')

def calculer_db():
    '''Valeur dB SPL maximale.'''
    y, sr = librosa.load(resource_path(r'data\out.wav'))
    max_amplitude = np.max(np.abs(y))
    if max_amplitude > 0:
        ref_pressure = 20e-6
        spl = 20 * np.log10(max_amplitude / ref_pressure)
    else:
        spl = 1
    return round(float(spl), 2)

def calculer_pourcentage_volume(db_level):
    percentage = 100.0
    reduction_dB = int(db_level - LIMIT_DB)
    if reduction_dB > 0:
        for _ in range(reduction_dB // 10):
            percentage *= 0.316
        remainder_dB = reduction_dB % 10
        if remainder_dB > 0:
            percentage *= 10 ** (-remainder_dB / 10)
    else:
        percentage = 99.0
    return int(percentage) / 100

def fade_volume(current_volume, target_volume, duration=1.0, steps=20):
    step_volume = (target_volume - current_volume) / steps
    step_duration = duration / steps
    for _ in range(steps):
        current_volume += step_volume
        volume = AudioUtilities.GetSpeakers()
        interface = volume.Activate(IAudioEndpointVolume._iid_, CLSCTX_ALL, None)
        volume_control = cast(interface, POINTER(IAudioEndpointVolume))
        volume_control.SetMasterVolumeLevelScalar(current_volume, None)
        time.sleep(step_duration)

def limit_volume(db_level, new_volume):
    devices = AudioUtilities.GetSpeakers()
    interface = devices.Activate(IAudioEndpointVolume._iid_, CLSCTX_ALL, None)
    volume = cast(interface, POINTER(IAudioEndpointVolume))
    current_volume = volume.GetMasterVolumeLevelScalar()
    fade_volume(current_volume, new_volume)
    print(f"Volume: {db_level:.2f} dB, Différence: {round(LIMIT_DB - db_level, 2)}, Volume: {new_volume}")

def run_audio_processing():
    global running
    comtypes.CoInitialize()
    try:
        while running:
            record_file()
            db_level = calculer_db()
            new_volume = calculer_pourcentage_volume(db_level)
            limit_volume(db_level, new_volume)
            time.sleep(WAIT_TIME)
    finally:
        comtypes.CoUninitialize()

def start_program():
    global running
    if not running:
        running = True
        start_button.config(state=tk.DISABLED)
        stop_button.config(state=tk.NORMAL)
        threading.Thread(target=run_audio_processing, daemon=True).start()

def stop_program():
    global running
    running = False
    start_button.config(state=tk.NORMAL)
    stop_button.config(state=tk.DISABLED)

def update_limit():
    global LIMIT_DB
    try:
        new_limit = float(limit_entry.get())
        LIMIT_DB = new_limit
        save_settings()
        messagebox.showinfo("Limite mise à jour", f"La nouvelle limite est maintenant fixée à {LIMIT_DB} dB.")
    except ValueError:
        messagebox.showerror("Erreur", "Veuillez entrer un nombre valide.")

def update_wait_time():
    global WAIT_TIME
    try:
        new_wait_time = int(wait_time_entry.get())
        WAIT_TIME = new_wait_time
        save_settings()
        messagebox.showinfo("Durée d'attente mise à jour", f"La nouvelle durée d'attente est maintenant fixée à {WAIT_TIME} secondes.")
    except ValueError:
        messagebox.showerror("Erreur", "Veuillez entrer un nombre valide.")

def reset_limit():
    """Réinitialise la limite dB à la valeur par défaut."""
    global LIMIT_DB
    LIMIT_DB = DEFAULT_LIMIT_DB
    limit_entry.delete(0, tk.END)
    limit_entry.insert(0, str(LIMIT_DB))
    save_settings()
    messagebox.showinfo("Réinitialisation", f"La limite dB a été réinitialisée à {LIMIT_DB} dB.")

def reset_wait_time():
    """Réinitialise la durée d'attente à la valeur par défaut."""
    global WAIT_TIME
    WAIT_TIME = DEFAULT_WAIT_TIME
    wait_time_entry.delete(0, tk.END)
    wait_time_entry.insert(0, str(WAIT_TIME))
    save_settings()
    messagebox.showinfo("Réinitialisation", f"La durée d'attente a été réinitialisée à {WAIT_TIME} secondes.")

# Création de l'interface Tkinter
root = tk.Tk()
root.title(f"Sound Limiter - {VERSION}")

# Charger les paramètres au démarrage
load_settings()

# Définir l'icône de la fenêtre
icon_path = resource_path(r"data\app.ico")
root.iconbitmap(icon_path)

# Ajouter le logo
logo_path = resource_path(r"data\app.png")
logo = tk.PhotoImage(file=logo_path).subsample(2, 2)

# Créer un cadre pour centrer le logo
logo_frame = tk.Frame(root)
logo_frame.pack(pady=10)

logo_label = tk.Label(logo_frame, image=logo)
logo_label.pack()

root.geometry("400x600")

# Centrer la fenêtre sur l'écran
window_width = 400
window_height = 600
screen_width = root.winfo_screenwidth()
screen_height = root.winfo_screenheight()
x = (screen_width // 2) - (window_width // 2)
y = (screen_height // 2) - (window_height // 2)
root.geometry(f"{window_width}x{window_height}+{x}+{y}")

start_button = tk.Button(root, text="Démarrer", command=start_program)
start_button.pack(pady=10)

stop_button = tk.Button(root, text="Arrêter", command=stop_program, state=tk.DISABLED)
stop_button.pack(pady=10)

# Limite dB
limit_label = tk.Label(root, text="Limite dB:")
limit_label.pack(pady=5)

limit_entry = tk.Entry(root, justify='center')
limit_entry.insert(0, str(LIMIT_DB))
limit_entry.pack(pady=5)

# Cadre pour les boutons de mise à jour et de réinitialisation
limit_buttons_frame = tk.Frame(root)
limit_buttons_frame.pack(pady=10)

update_limit_button = tk.Button(limit_buttons_frame, text="Mettre à jour la limite", command=update_limit)
update_limit_button.pack(side=tk.LEFT)

reset_limit_button = tk.Button(limit_buttons_frame, text="Réinitialiser la limite", command=reset_limit)
reset_limit_button.pack(side=tk.LEFT)

# Durée d'attente
wait_time_label = tk.Label(root, text="Durée d'attente (sec):")
wait_time_label.pack(pady=5)

wait_time_entry = tk.Entry(root, justify='center')
wait_time_entry.insert(0, str(WAIT_TIME))
wait_time_entry.pack(pady=5)

# Cadre pour les boutons de mise à jour et de réinitialisation
wait_time_buttons_frame = tk.Frame(root)
wait_time_buttons_frame.pack(pady=10)

update_wait_time_button = tk.Button(wait_time_buttons_frame, text="Mettre à jour la durée", command=update_wait_time)
update_wait_time_button.pack(side=tk.LEFT)

reset_wait_time_button = tk.Button(wait_time_buttons_frame, text="Réinitialiser la durée", command=reset_wait_time)
reset_wait_time_button.pack(side=tk.LEFT)

# Créer l'icône pour la systray
def create_systray_icon():
    global icon
    icon_image = Image.open(resource_path(r"data\app.png"))
    icon = pystray.Icon("sound_limiter", icon_image, "Sound Limiter", menu=pystray.Menu(
        pystray.MenuItem("Ouvrir", show_window),
        pystray.MenuItem("Quitter", quit_program)
    ))

def show_window(icon, item):
    root.deiconify()
    icon.stop()

def quit_program(icon, item):
    stop_program()
    save_settings()
    icon.stop()
    root.quit()

# Gérer la fermeture de la fenêtre
def on_closing():
    root.withdraw()
    create_systray_icon()
    icon.run()

root.protocol("WM_DELETE_WINDOW", on_closing)

root.mainloop()
